/***********************************************************************
 * Time-stamp: <Mittwoch, 09.05.2001, 14:31:36; edited by heiko>
 * 
 * Tina's random number generators
 *
 * seral testing for random number generators
 *
 * Copyright by Heiko Bauke, Magdeburg 2001 
 *
 * heiko.bauke@student.uni-magdeburg.de
 *
 * This is free software. You may use and distribute this software under
 * the terms of the GNU GENERAL PUBLIC LICENSE Version 2. See COPYING
 * for details.
 *
 ***********************************************************************/

#include <cstdlib>
#include <cmath>
#include <iostream>
#include <iomanip>
#include "trng.h"

using namespace TRNG;
using namespace std;


void serial_test(RNG &r, long runs, long n, long num_classes) {
  // r: random number generator
  // bit: bit to test
  // runs: number of Chi-Square tests
  // n: random numbers per test
  // num_classes: number of classes for equidistribution test
  vector<double> observ, prob;
  long i, num_fail;
  double Chi2, Chi2P;
  cout.setf(ios::fixed);
  observ.resize(num_classes*num_classes);
  prob.resize(num_classes*num_classes, 
	      1.0/num_classes/num_classes);
  num_fail=0l;
  while (runs>0l) {
    for (i=0; i<num_classes*num_classes; ++i)
      observ[i]=0.0;
    for (i=0l; i<n; ++i)
      ++observ[static_cast<long>(r.uniform()*num_classes)*num_classes
	      +static_cast<long>(r.uniform()*num_classes)];
    Chi2=Chisquare(prob, observ);
    Chi2P=Chisquare_prob(Chi2, num_classes*num_classes-1l);
    if (Chi2P<0.01 || 0.99<Chi2P) {
#ifdef LaTeX
      cout << "\\textbf{";
#endif
      ++num_fail;
    }
    cout << setfill(' ') << setprecision(3) << Chi2P;
#ifdef LaTeX
    if (Chi2P<0.01 || 0.99<Chi2P)
      cout << "}";
    cout << " & ";
#else
    cout << '\t';
#endif
    --runs;
  }
  if (num_fail>2l)
#ifdef LaTeX
    cout << "\\textbf{durchgefallen}";
#else
    cout << "failed";
#endif
  else
    if (num_fail>1l) 
#ifdef LaTeX
      cout << "\\textbf{suspekt}";
#else
      cout << "suspect";
#endif
    else
#ifdef LaTeX
      cout << "o.\\,k.";
#else
      cout << "o.k.";
#endif;
#ifdef LaTeX
  cout << "\\\\" << endl;
#else
  cout << endl;
#endif
}


void serial_test_main(RNG &r, long runs, long sub_max, long n, long num_classes) {
  long j;
#ifdef LaTeX
  cout << "\\multiciol{" << runs+1l <<"}{@{}l@{}}{Generator " << r.name() 
       << "}\\\\" << endl << "\\midrule" << endl;
#else
  cout << "Generator " << r.name() << endl;
#endif
  for (j=1l; j<=sub_max; ++j) {
#ifdef LaTeX
    cout << "\\multiciol{" << runs+2l <<"}{@{}l@{}}{Splitting level " << j 
	 << "}\\\\" << endl;
#else
    cout << "Splitting level " << j << endl;
#endif
    r.reset();
    r.split(j, 0l);
    serial_test(r, runs, n, num_classes);
  }
}


int main(int argc, char *argv[]) {
  if (argc!=5) {
    cerr << 
      "Tina's Random Number Generator Library\n";
    cerr << 
      "(P) & (C) by Heiko Bauke, Magdeburg 2001\n\n";
    cerr << 
      "Serial test\n";
    cerr << 
      "-----------\n\n";
    cerr << 
      "synopsis:\n";
    cerr << 
      "$ " << argv[0] << " runs sub_max n num_classes\n\n";
    cerr <<
      "usefull arguments: runs=5; sub_max=128; n=5000; num_classes=12; n>num_classes^2\n";
    cerr <<
      "The Programm counts how often pairs (p1, p2) (0<=p_i<num_classes)\n";
    cerr <<  
      "appear and makes a Chi^2-test.\n\n";
    cerr 
      << "Four further informations see Donald E. Knuth\n";
    cerr << "\"The Art of Computer Programming\", 1981\n";
  } else {
    long runs, sub_max, n, num_classes;
    runs=atol(argv[1]);
    sub_max=atol(argv[2]);
    n=atol(argv[3]);
    num_classes=atol(argv[4]);
#ifdef LaTeX
    cout << num_classes << " Klassen \\\\" << endl;
    cout << n << " W\\\"urfe je Test\\\\" << endl;
#else
    cout << num_classes << " classes" << endl;
    cout << n << " throws per test" << endl;
#endif
    ParkMiller r1;
    RAND r2;
    LCG64 r3;
    LMC2 r4;
    LMC3 r5;
    LMC4 r6;
    CLCG2 r7;
    CLCG3 r8;
    CLCG4 r9;
    EINV r10;
    EINVLCG64 r11;
    serial_test_main(r1, runs, sub_max, n, num_classes);
    serial_test_main(r2, runs, sub_max, n, num_classes);
    serial_test_main(r3, runs, sub_max, n, num_classes);
    serial_test_main(r4, runs, sub_max, n, num_classes);
    serial_test_main(r5, runs, sub_max, n, num_classes);
    serial_test_main(r6, runs, sub_max, n, num_classes);
    serial_test_main(r7, runs, sub_max, n, num_classes);
    serial_test_main(r8, runs, sub_max, n, num_classes);
    serial_test_main(r9, runs, sub_max, n, num_classes);
    serial_test_main(r10, runs, sub_max, n, num_classes);
    serial_test_main(r11, runs, sub_max, n, num_classes);
  }
  return(EXIT_SUCCESS);
}
